#!/usr/bin/env python
#
# Copyright (c) 2006-2007 Ricardo Garcia Gonzalez
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# 
# Except as contained in this notice, the name(s) of the above copyright
# holders shall not be used in advertising or otherwise to promote the
# sale, use or other dealings in this Software without prior written
# authorization.
#
import datetime
import httplib
import math
import optparse
import os
import random
import re
import socket
import sys
import string
import time
import urllib2

# Global constants
const_video_url_re = re.compile(r'(?:http://)?(?:www\.)?pornotube\.com/media\.php\?m=(\d+)')
const_normalized_url_str = 'http://www.pornotube.com/media.php?m=%s'
const_video_url_age_post_str = 'bMonth=%s&bDay=%s&bYear=%s&verifyAge=true&submit=CONTINUE+%%C2%%BB'
const_video_url_id_re = re.compile(r'"player/v\.swf\?v=([^"]+)')
const_video_title_re = re.compile(r'<span>Viewing Video: </span><span class="blue">([^<]*)</span>', re.M | re.I)
const_video_url_player_str = 'http://www.pornotube.com/player/player.php?%s'
const_video_url_mid_re = re.compile(r'&mediaId=(\d+)&')
const_video_url_uid_re = re.compile(r'&userId=(\d+)&')
const_video_url_mdomain_re = re.compile(r'&mediaDomain=([^&]+)&')
const_video_url_real_str = 'http://%s.pornotube.com/%s/%s.flv'
const_1k = 1024
const_initial_block_size = 10 * const_1k
const_epsilon = 0.0001
const_max_years = 50
const_min_years = 25
const_min_month = 1
const_max_month = 12
const_min_day = 1
const_max_day = 28

# Fill age confirmation form data
def get_age_post_data():
	global const_video_url_age_post_str
	global const_max_years
	global const_min_years
	global const_min_month
	global const_max_month
	global const_min_day
	global const_max_day

	today = datetime.date.today()
	year = random.randint(today.year - const_max_years, today.year - const_min_years)
	month = '%02d' % random.randint(const_min_month, const_max_month) # Month is zero-padded
	day = random.randint(const_min_day, const_max_day)
	return const_video_url_age_post_str % (month, day, year)

# Print error message, followed by standard advice information, and then exit
def error_advice_exit(error_text):
	sys.stderr.write('Error: %s.\n' % error_text)
	sys.stderr.write('Try again several times. It may be a temporary problem.\n')
	sys.stderr.write('Other typical problems:\n\n')
	sys.stderr.write('* URL is wrong.\n')
	sys.stderr.write('* Content is not a flash video.\n')
	sys.stderr.write('* Video no longer exists.\n')
	sys.stderr.write('* The connection was cut suddenly for some reason.\n')
	sys.stderr.write('* pornotube changed their system, and the program no longer works.\n')
	sys.stderr.write('\nTry to confirm you are able to view the video using a web browser.\n')
	sys.stderr.write('When using a proxy, make sure http_proxy has http://host:port format.\n')
	sys.stderr.write('Try again several times and contact me if the problem persists.\n')
	sys.exit('\n')

# Wrapper to create custom requests with typical headers
def request_create(url, data=None):
	retval = urllib2.Request(url)
	if data is not None:
		retval.add_data(data)
	# Try to mimic Firefox, at least a little bit
	retval.add_header('User-Agent', 'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1) Gecko/20061010 Firefox/2.0')
	retval.add_header('Accept-Charset', 'ISO-8859-1,utf-8;q=0.7,*;q=0.7')
	retval.add_header('Accept', 'text/xml,application/xml,application/xhtml+xml,text/html;q=0.9,text/plain;q=0.8,image/png,*/*;q=0.5')
	retval.add_header('Accept-Language', 'en-us,en;q=0.5')
	return retval

# Perform a request, process headers and return response
def perform_request(url, data=None):
	request = request_create(url, data)
	response = urllib2.urlopen(request)
	return response

# Conditional print
def cond_print(str):
	global cmdl_opts
	if not (cmdl_opts.quiet or cmdl_opts.get_url):
		sys.stdout.write(str)
		sys.stdout.flush()

# Title string normalization
def title_string_norm(title):
	title = ''.join((x in string.ascii_letters or x in string.digits) and x or ' ' for x in title)
	title = '_'.join(title.split())
	title = title.lower()
	return title

# Generic download step
def download_step(return_data_flag, step_title, step_error, url, post_data=None):
	try:
		cond_print('%s... ' % step_title)
		data = perform_request(url, post_data).read()
		cond_print('done.\n')
		if return_data_flag:
			return data
		return None

	except (urllib2.URLError, ValueError, httplib.HTTPException, TypeError, socket.error):
		cond_print('failed.\n')
		error_advice_exit(step_error)

	except KeyboardInterrupt:
		sys.exit('\n')

# Generic extract step
def extract_step(step_title, step_error, regexp, data):
	try:
		cond_print('%s... ' % step_title)
		match = regexp.search(data)
		
		if match is None:
			cond_print('failed.\n')
			error_advice_exit(step_error)
		
		extracted_data = match.group(1)
		cond_print('done.\n')
		return extracted_data
	
	except KeyboardInterrupt:
		sys.exit('\n')

# Calculate new block size based on previous block size
def new_block_size(before, after, bytes):
	new_min = max(bytes / 2.0, 1.0)
	new_max = max(bytes * 2.0, 1.0)
	dif = after - before
	if dif < const_epsilon:
		return int(new_max)
	rate = bytes / dif
	if rate > new_max:
		return int(new_max)
	if rate < new_min:
		return int(new_min)
	return int(rate)

# Get optimum 1k exponent to represent a number of bytes
def optimum_k_exp(num_bytes):
	global const_1k
	if num_bytes == 0:
		return 0
	return long(math.log(num_bytes, const_1k))

# Get optimum representation of number of bytes
def format_bytes(num_bytes):
	global const_1k
	try:
		exp = optimum_k_exp(num_bytes)
		suffix = 'bkMGTPEZY'[exp]
		if exp == 0:
			return '%s%s' % (num_bytes, suffix)
		converted = float(num_bytes) / float(const_1k**exp)
		return '%.2f%s' % (converted, suffix)
	except IndexError:
		sys.exit('Error: internal error formatting number of bytes.')

# Calculate ETA and return it in string format as MM:SS
def calc_eta(start, now, total, current):
	dif = now - start
	if current == 0 or dif < const_epsilon:
		return '--:--'
	rate = float(current) / dif
	eta = long((total - current) / rate)
	eta_mins = eta / 60
	eta_secs = eta % 60
	if eta_mins > 99:
		return '--:--'
	return '%02d:%02d' % (eta_mins, eta_secs)

# Calculate speed and return it in string format
def calc_speed(start, now, bytes):
	dif = now - start
	if bytes == 0 or dif < const_epsilon:
		return 'N/A b'
	return format_bytes(float(bytes) / dif)

# Create the command line options parser and parse command line
cmdl_usage = 'usage: %prog [options] video_url'
cmdl_version = '2007.10.09'
cmdl_parser = optparse.OptionParser(usage=cmdl_usage, version=cmdl_version, conflict_handler='resolve')
cmdl_parser.add_option('-h', '--help', action='help', help='print this help text and exit')
cmdl_parser.add_option('-v', '--version', action='version', help='print program version and exit')
cmdl_parser.add_option('-o', '--output', dest='outfile', metavar='FILE', help='output video file name')
cmdl_parser.add_option('-q', '--quiet', action='store_true', dest='quiet', help='activates quiet mode')
cmdl_parser.add_option('-s', '--simulate', action='store_true', dest='simulate', help='do not download video')
cmdl_parser.add_option('-t', '--title', action='store_true', dest='use_title', help='use title in file name')
cmdl_parser.add_option('-g', '--get-url', action='store_true', dest='get_url', help='print final video URL only')
(cmdl_opts, cmdl_args) = cmdl_parser.parse_args()

# Get video URL
if len(cmdl_args) != 1:
	cmdl_parser.print_help()
	sys.exit('\n')
video_url = cmdl_args[0]

# Verify video URL format and extract URL data to normalize URL
video_url_mo = const_video_url_re.match(video_url)
if video_url_mo is None:
	sys.exit('Error: URL does not seem to be a pornotube video URL. If it is, report a bug.')
video_url_id = video_url_mo.group(1)
video_url = const_normalized_url_str % video_url_id

# Check conflicting options
if cmdl_opts.outfile is not None and (cmdl_opts.simulate or cmdl_opts.get_url):
	sys.stderr.write('Warning: video file name given but will not be used.\n')

if cmdl_opts.outfile is not None and cmdl_opts.use_title:
	sys.exit('Error: using the video title conflicts with using a given file name.')

if cmdl_opts.quiet and cmdl_opts.get_url:
		sys.exit('Error: cannot be quiet and print final URL at the same time.')

# Get output file name 
if cmdl_opts.outfile is None:
	video_filename = '%s.flv' % video_url_id
else:
	video_filename = cmdl_opts.outfile

# Check name
if not video_filename.lower().endswith('.flv'):
	sys.stderr.write('Warning: video file name does not end in .flv\n')

# Test writable file
if not (cmdl_opts.simulate or cmdl_opts.get_url):
	try:
		disk_test = open(video_filename, 'wb')
		disk_test.close()

	except (OSError, IOError):
		sys.exit('Error: unable to open %s for writing.' % video_filename)

# Install cookie and proxy handlers
urllib2.install_opener(urllib2.build_opener(urllib2.ProxyHandler()))
urllib2.install_opener(urllib2.build_opener(urllib2.HTTPCookieProcessor()))

# Retrieve video webpage
video_webpage = download_step(True, 'Sending age and retrieving video webpage', 'unable to retrieve video webpage', video_url, get_age_post_data())

# Extract video title if needed
if cmdl_opts.use_title:
	video_title = extract_step('Extracting video title', 'unable to extract video title', const_video_title_re, video_webpage)

# Extract needed player URL parameters
video_id_param = extract_step('Extracting player URL parameters', 'unable to extract player URL parameters', const_video_url_id_re, video_webpage)
video_url_player = const_video_url_player_str % video_id_param

# Retrieve video player data
video_url_player_params = download_step(True, 'Retrieving video player parameters', 'unable to retrieve video player parameters', video_url_player)

# Extract real video URL data
video_url_mid = extract_step('Extracting mid', 'unable to extract mid', const_video_url_mid_re, video_url_player_params)
video_url_uid = extract_step('Extracting uid', 'unable to extract uid', const_video_url_uid_re, video_url_player_params)
video_url_mdomain = extract_step('Extracting mdomain', 'unable to extract mdomain', const_video_url_mdomain_re, video_url_player_params)
video_url_real = const_video_url_real_str % (video_url_mdomain, video_url_uid, video_url_mid)

# Retrieve video data
try:
	cond_print('Requesting video file... ')
	video_data = perform_request(video_url_real)
	cond_print('done\n')
	cond_print('Video data found at %s\n' % video_data.geturl())

	if cmdl_opts.get_url:
		print video_data.geturl()

	if cmdl_opts.simulate or cmdl_opts.get_url:
		sys.exit()

	video_file = open(video_filename, 'wb')
	try:
		video_len = long(video_data.info()['Content-length'])
		video_len_str = format_bytes(video_len)
	except KeyError:
		video_len = None
		video_len_str = 'N/A'

	byte_counter = 0
	block_size = const_initial_block_size
	start_time = time.time()
	while True:
		if video_len is not None:
			percent = float(byte_counter) / float(video_len) * 100.0
			percent_str = '%.1f' % percent
			eta_str = calc_eta(start_time, time.time(), video_len, byte_counter)
		else:
			percent_str = '---.-'
			eta_str = '--:--'
		counter = format_bytes(byte_counter)
		speed_str = calc_speed(start_time, time.time(), byte_counter)
		cond_print('\rRetrieving video data: %5s%% (%8s of %s) at %8s/s ETA %s ' % (percent_str, counter, video_len_str, speed_str, eta_str))

		before = time.time()
		video_block = video_data.read(block_size)
		after = time.time()
		dl_bytes = len(video_block)
		if dl_bytes == 0:
			break
		byte_counter += dl_bytes
		video_file.write(video_block)
		block_size = new_block_size(before, after, dl_bytes)

	if video_len is not None and byte_counter != video_len:
		error_advice_exit('server did not send the expected ammount of data')

	video_file.close()
	cond_print('done.\n')
	cond_print('Video data saved to %s\n' % video_filename)

except (urllib2.URLError, ValueError, httplib.HTTPException, TypeError, socket.error):
	cond_print('failed.\n')
	error_advice_exit('unable to download video data')

except KeyboardInterrupt:
	sys.exit('\n')

# Rename video file if needed
if cmdl_opts.use_title:
	try:
		final_filename = '%s-%s.flv' % (title_string_norm(video_title), video_url_id)
		os.rename(video_filename, final_filename)
		cond_print('Video file renamed to %s\n' % final_filename)
	
	except OSError:
		sys.stderr.write('Warning: unable to rename file.\n')

	except KeyboardInterrupt:
		sys.exit('\n')

# Finish
sys.exit()
